//
// This fragment shader defines a reference implementation for Physically Based Shading of
// a microfacet surface material defined by a glTF model.
//
// References:
// [1] Real Shading in Unreal Engine 4
//     http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf
// [2] Physically Based Shading at Disney
//     http://blog.selfshadow.com/publications/s2012-shading-course/burley/s2012_pbs_disney_brdf_notes_v3.pdf
// [3] README.md - Environment Maps
//     https://github.com/KhronosGroup/glTF-WebGL-PBR/#environment-maps
// [4] "An Inexpensive BRDF Model for Physically based Rendering" by Christophe Schlick
//     https://www.cs.virginia.edu/~jdl/bib/appearance/analytic%20models/schlick94b.pdf

#if defined(BUMPED) && defined(USE_IBL)
varying vec3 v_cameraDirectionWorldSpace;
varying mat3 v_tangentToWorldSpaceTransformMatrix;
#endif

uniform vec4  u_albedo;
#if defined(HAS_ALBEDOMAP)
uniform sampler2D u_albedoMap;
#endif

#if defined(PACKED_METALLIC_ROUGHNESS)
uniform vec2  u_metallicRoughness; // 与glTF保持一致
#if defined(HAS_METALLICROUGHNESSMAP)
uniform sampler2D u_metallicRoughnessMap;
#endif //HAS_METALLICROUGHNESSMAP
#else //PACKED_METALLIC_ROUGHNESS
uniform float  u_metallic;
uniform float  u_roughness;
#if defined(HAS_METALLICROUGHNESSMAP)
uniform sampler2D u_metallicMap;
uniform sampler2D u_roughnessMap;
#endif //HAS_METALLICROUGHNESSMAP
#endif //PACKED_METALLIC_ROUGHNESS

#if defined(HAS_AOMAP)
uniform sampler2D u_aoMap;
uniform float u_ao;
#endif

#if defined(HAS_EMISSIVEMAP)
uniform sampler2D u_emissiveMap;
uniform vec3 u_emissiveFactor;
#endif

#ifdef USE_IBL
uniform samplerCube u_diffuseEnvTexture;
uniform samplerCube u_specularEnvTexture;
#endif
uniform sampler2D u_brdfLUT;

#if defined(DEBUG)
// debugging flags used for shader output of intermediate PBR variables
uniform vec4 u_scaleDiffBaseMR; // [diffusion, baseColor, metallic, roughness], default [0.0, 0.0, 0.0, 0.0]
uniform vec4 u_scaleFGDSpec;    // [F, G, D, specular], default [0.0, 0.0, 0.0, 0.0]
uniform vec4 u_scaleIBLAmbient; // [IBLDiffScale, IBLSpecScale, 0, 0], default [1.0, 1.0, 1.0, 1.0]
vec3 g_F = vec3(0.0, 0.0, 0.0);
float g_G = 0.0;
float g_D = 0.0;
vec3 g_diffuseContrib = vec3(0.0, 0.0, 0.0);
vec3 g_specContrib = vec3(0.0, 0.0, 0.0);
#endif

// Encapsulate the various inputs used by the various functions in the shading equation
// We store values in this struct to simplify the integration of alternative implementations
// of the shading terms, outlined in the Readme.MD Appendix.
struct PBRInfo
{
    float NdotL;                  // cos angle between normal and light direction
    float NdotV;                  // cos angle between normal and view direction
    float NdotH;                  // cos angle between normal and half vector
    float LdotH;                  // cos angle between light direction and half vector
    float VdotH;                  // cos angle between view direction and half vector
    float perceptualRoughness;    // roughness value, as authored by the model creator (input to shader)
    float metalness;              // metallic value at the surface
    vec3 reflectance0;            // full reflectance color (normal incidence angle)
    vec3 reflectance90;           // reflectance color at grazing angle
    float alphaRoughness;         // roughness mapped to a more linear change in the roughness (proposed by [2])
    vec3 diffuseColor;            // color contribution from diffuse lighting
    vec3 specularColor;           // color contribution from specular lighting
};

const float M_PI = 3.141592653589793;
const float c_minRoughness = 0.004;

vec4 SRGBtoLINEAR(vec4 srgbIn)
{
#ifdef MANUAL_SRGB
#ifdef SRGB_FAST_APPROXIMATION
    vec3 linOut = pow(srgbIn.xyz,vec3(2.2));
#else //SRGB_FAST_APPROXIMATION
    vec3 bLess = step(vec3(0.04045),srgbIn.xyz);
    vec3 linOut = mix( srgbIn.xyz/vec3(12.92), pow((srgbIn.xyz+vec3(0.055))/vec3(1.055),vec3(2.4)), bLess );
#endif //SRGB_FAST_APPROXIMATION
    return vec4(linOut,srgbIn.w);;
#else //MANUAL_SRGB
    return srgbIn;
#endif //MANUAL_SRGB
}

vec3 adjustedTextureCubeCoord(vec3 coord)
{
    // 立方体采样有问题，需要对yz取反
    return vec3(coord.x, -coord.yz);
}

// Calculation of the lighting contribution from an optional Image Based Light source.
// Precomputed Environment Maps are required uniform inputs and are computed as outlined in [1].
// See our README.md on Environment Maps [3] for additional discussion.
#ifdef USE_IBL
vec3 getIBLContribution(float perceptualRoughness, float NdotV, vec3 diffuseColor, vec3 specularColor, vec3 n, vec3 reflection)
{
    // retrieve a scale and bias to F0. See [1], Figure 3
    // 这里与原始的不同，直接采用roughness作为y方向，并且brdfLUT已经是在线性空间了，不再需要做处理
    vec3 brdf = texture2D(u_brdfLUT, vec2(NdotV, perceptualRoughness)).rgb;
    vec3 diffuseLight = SRGBtoLINEAR(textureCube(u_diffuseEnvTexture, adjustedTextureCubeCoord(n))).rgb;
    
#ifdef USE_TEX_LOD
    float mipCount = 9.0; // resolution of 512x512
    float lod = (perceptualRoughness * mipCount);
    vec3 specularLight = SRGBtoLINEAR(textureCubeLod(u_specularEnvTexture, reflection, lod)).rgb;
#else
    // 这里与原始的不同，brdfLUT已经是在线性空间了，不再需要做处理
    vec3 specularLight = SRGBtoLINEAR(textureCube(u_specularEnvTexture, adjustedTextureCubeCoord(reflection))).rgb;
#endif
    
    vec3 diffuse = diffuseLight * diffuseColor;
    vec3 specular = specularLight * (specularColor * brdf.x + brdf.y);

#if defined(DEBUG)
    // For presentation, this allows us to disable IBL terms
    diffuse *= u_scaleIBLAmbient.x;
    specular *= u_scaleIBLAmbient.y;
#endif
    
    return diffuse + specular;
}
#else
vec3 getAmbientContribution(float perceptualRoughness, float NdotV, vec3 diffuseColor, vec3 specularColor)
{
    // retrieve a scale and bias to F0. See [1], Figure 3
    // 这里与原始的不同，直接采用roughness作为y方向，并且brdfLUT已经是在线性空间了，不再需要做处理
    vec3 brdf = texture2D(u_brdfLUT, vec2(NdotV, perceptualRoughness)).rgb;
    vec3 diffuseLight = u_ambientColor.rgb * NdotV;// 这里根据方向来处理效果非常好
    vec3 specularLight = u_ambientColor.rgb * NdotV * (1.0 - perceptualRoughness);
    
    vec3 diffuse = diffuseLight * diffuseColor;
    vec3 specular = specularLight * (specularColor * brdf.x + brdf.y);

    return diffuse + specular;
}
#endif

// Basic Lambertian diffuse
// Implementation from Lambert's Photometria https://archive.org/details/lambertsphotome00lambgoog
// See also [1], Equation 1
vec3 diffuse(PBRInfo pbrInputs)
{
    return pbrInputs.diffuseColor / M_PI;
}

// The following equation models the Fresnel reflectance term of the spec equation (aka F())
// Implementation of fresnel from [4], Equation 15
vec3 specularReflection(PBRInfo pbrInputs)
{
    return pbrInputs.reflectance0 + (pbrInputs.reflectance90 - pbrInputs.reflectance0) * pow(clamp(1.0 - pbrInputs.VdotH, 0.0, 1.0), 5.0);
}

// This calculates the specular geometric attenuation (aka G()),
// where rougher material will reflect less light back to the viewer.
// This implementation is based on [1] Equation 4, and we adopt their modifications to
// alphaRoughness as input as originally proposed in [2].
float geometricOcclusion(PBRInfo pbrInputs)
{
    float NdotL = pbrInputs.NdotL;
    float NdotV = pbrInputs.NdotV;
    float r = pbrInputs.alphaRoughness;
    
    float attenuationL = 2.0 * NdotL / (NdotL + sqrt(r * r + (1.0 - r * r) * (NdotL * NdotL)));
    float attenuationV = 2.0 * NdotV / (NdotV + sqrt(r * r + (1.0 - r * r) * (NdotV * NdotV)));
    return attenuationL * attenuationV;
}

// The following equation(s) model the distribution of microfacet normals across the area being drawn (aka D())
// Implementation from "Average Irregularity Representation of a Roughened Surface for Ray Reflection" by T. S. Trowbridge, and K. P. Reitz
// Follows the distribution function recommended in the SIGGRAPH 2013 course notes from EPIC Games [1], Equation 3.
float microfacetDistribution(PBRInfo pbrInputs)
{
    float roughnessSq = pbrInputs.alphaRoughness * pbrInputs.alphaRoughness;
    float f = (pbrInputs.NdotH * roughnessSq - pbrInputs.NdotH) * pbrInputs.NdotH + 1.0;
    return roughnessSq / (M_PI * f * f);
}

vec3 getLightContribution(vec3 l, vec3 n, vec3 v, float NdotV, vec3 diffuseColor, vec3 specularColor, vec3 lightColor, float metallic, float perceptualRoughness, vec3 specularEnvironmentR0, vec3 specularEnvironmentR90, float alphaRoughness)
{
    vec3 h = normalize(l+v);                          // Half vector between both l and v

    float NdotL = clamp(dot(n, l), 0.001, 1.0);
    float NdotH = clamp(dot(n, h), 0.0, 1.0);
    float LdotH = clamp(dot(l, h), 0.0, 1.0);
    float VdotH = clamp(dot(v, h), 0.0, 1.0);
    
    PBRInfo pbrInputs = PBRInfo(
                                NdotL,
                                NdotV,
                                NdotH,
                                LdotH,
                                VdotH,
                                perceptualRoughness,
                                metallic,
                                specularEnvironmentR0,
                                specularEnvironmentR90,
                                alphaRoughness,
                                diffuseColor,
                                specularColor
                                );
    
    // Calculate the shading terms for the microfacet specular shading model
    vec3 F = specularReflection(pbrInputs);
    float G = geometricOcclusion(pbrInputs);
    float D = microfacetDistribution(pbrInputs);
    
    // Calculation of analytical lighting contribution
    vec3 diffuseContrib = (1.0 - F) * diffuse(pbrInputs);
    vec3 specContrib = F * G * D / (4.0 * NdotL * NdotV);
#if defined(DEBUG)
    g_F += F;
    g_G += G;
    g_D += D;
    g_diffuseContrib += diffuseContrib;
    g_specContrib += specContrib;
#endif
    // Obtain final intensity as reflectance (BRDF) scaled by the energy of the light (cosine law)
    vec3 color = NdotL * lightColor * (diffuseContrib + specContrib);
    
    return color;
}

vec4 computeLighting(vec3 normalVector)
{
    // Metallic and Roughness material properties are packed together
    // In glTF, these factors can be specified by fixed scalar values
    // or from a metallic-roughness map
#if defined(PACKED_METALLIC_ROUGHNESS)
    float perceptualRoughness = u_metallicRoughness.y;
    float metallic = u_metallicRoughness.x;
#ifdef HAS_METALLICROUGHNESSMAP
    // Roughness is stored in the 'g' channel, metallic is stored in the 'b' channel.
    // This layout intentionally reserves the 'r' channel for (optional) occlusion map data
    vec4 mrSample = SRGBtoLINEAR(texture2D(u_metallicRoughnessMap, v_texCoord));
    perceptualRoughness = mrSample.g * perceptualRoughness;
    metallic = mrSample.b * metallic;
#endif // HAS_METALLICROUGHNESSMAP
#else // PACKED_METALLIC_ROUGHNESS
    float perceptualRoughness = u_roughness;
    float metallic = u_metallic;
#ifdef HAS_METALLICROUGHNESSMAP
    metallic = SRGBtoLINEAR(texture2D(u_metallicMap, v_texCoord)).r * metallic;
    perceptualRoughness = SRGBtoLINEAR(texture2D(u_roughnessMap, v_texCoord)).r * perceptualRoughness;
#endif // HAS_METALLICROUGHNESSMAP
#endif // PACKED_METALLIC_ROUGHNESS
    // GGX with roughtness to 0 would mean no specular at all, using max(roughness, 0.004) here to match HDrenderloop roughtness remapping.
    perceptualRoughness = clamp(perceptualRoughness, c_minRoughness, 1.0);
    metallic = clamp(metallic, 0.0, 1.0);
    // Roughness is authored as perceptual roughness; as is convention,
    // convert to material roughness by squaring the perceptual roughness [2].
    float alphaRoughness = perceptualRoughness * perceptualRoughness;
    
    // The albedo may be defined from a base texture or a flat color
#ifdef HAS_ALBEDOMAP
    vec4 baseColor = SRGBtoLINEAR(texture2D(u_albedoMap, v_texCoord)) * u_albedo;
#else
    vec4 baseColor = u_albedo;
#endif
    
    vec3 f0 = vec3(0.04);
    vec3 diffuseColor = baseColor.rgb * (vec3(1.0) - f0);
    diffuseColor *= 1.0 - metallic;
    vec3 specularColor = mix(f0, baseColor.rgb, metallic);
    
    // Compute reflectance.
    float reflectance = max(max(specularColor.r, specularColor.g), specularColor.b);
    
    // For typical incident reflectance range (between 4% to 100%) set the grazing reflectance to 100% for typical fresnel effect.
    // For very low reflectance range on highly diffuse objects (below 4%), incrementally reduce grazing reflecance to 0%.
    float reflectance90 = clamp(reflectance * 25.0, 0.0, 1.0);
    vec3 specularEnvironmentR0 = specularColor.rgb;
    vec3 specularEnvironmentR90 = vec3(1.0, 1.0, 1.0) * reflectance90;
    
    vec3 n = normalVector;                 // normal at surface point
    vec3 v = normalize(v_cameraDirection);            // Vector from surface point to camera
    float NdotV = clamp(abs(dot(n, v)), 0.001, 1.0);

    // reflectance equation
    vec3 color = vec3(0.0);
    // Directional light contribution
#if (DIRECTIONAL_LIGHT_COUNT > 0)
    for (int i = 0; i < DIRECTIONAL_LIGHT_COUNT; ++i)
    {
        // calculate per-light radiance
#if defined(BUMPED)
        vec3 lightDirection = normalize(v_directionalLightDirection[i] * 2.0);
#else
        vec3 lightDirection = normalize(u_directionalLightDirection[i] * 2.0);
#endif
        vec3 L = -lightDirection;
        vec3 radiance = u_directionalLightColor[i];
        
        color += getLightContribution(L, n, v, NdotV, diffuseColor, specularColor, radiance, metallic, perceptualRoughness, specularEnvironmentR0, specularEnvironmentR90, alphaRoughness);
    }
#endif
    
    // Point light contribution
#if (POINT_LIGHT_COUNT > 0)
    for (int i = 0; i < POINT_LIGHT_COUNT; ++i)
    {
        // calculate per-light radiance
        vec3 L = normalize(v_vertexToPointLightDirection[i]);
        float distance = length(v_vertexToPointLightDirection[i] * u_pointLightRangeInverse[i]);
        float attenuation = 1.0 / (distance * distance);
        vec3 radiance = u_pointLightColor[i] * attenuation;
        
        color += getLightContribution(L, n, v, NdotV, diffuseColor, specularColor, radiance, metallic, perceptualRoughness, specularEnvironmentR0, specularEnvironmentR90, alphaRoughness);
    }
#endif
    
    // Spot light contribution
#if (SPOT_LIGHT_COUNT > 0)
    for (int i = 0; i < SPOT_LIGHT_COUNT; ++i)
    {
        // calculate per-light radiance
        vec3 L = normalize(v_vertexToSpotLightDirection[i]);
        float distance = length(v_vertexToSpotLightDirection[i] * u_spotLightRangeInverse[i]);
        float attenuation = 1.0 / (distance * distance);
        
#if defined(BUMPED)
        vec3 spotLightDirection = normalize(v_spotLightDirection[i] * 2.0);
#else
        vec3 spotLightDirection = normalize(u_spotLightDirection[i] * 2.0);
#endif
        
        // "-lightDirection" is used because light direction points in opposite direction to spot direction.
        float spotCurrentAngleCos = dot(spotLightDirection, -L);
        
        // Apply spot attenuation
        attenuation *= smoothstep(u_spotLightOuterAngleCos[i], u_spotLightInnerAngleCos[i], spotCurrentAngleCos);
        vec3 radiance = u_spotLightColor[i] * attenuation;
        
        color += getLightContribution(L, n, v, NdotV, diffuseColor, specularColor, radiance, metallic, perceptualRoughness, specularEnvironmentR0, specularEnvironmentR90, alphaRoughness);
    }
#endif
    
    // Calculate lighting contribution from image based lighting source (IBL)
#ifdef USE_IBL
    // 这里比较特殊，需要在世界坐标系下计算，而上面的不影响是因为都是点乘，最后结果在哪个空间中都无所谓
    // 但这里的normal和reflection都需要在世界坐标系下，因为计算出来的纹理贴图处于世界坐标系
    // 为何不全部在世界坐标系下计算而留下一部分在切线坐标系是为了性能考虑，可以减少片段着色器下的矩阵运算
    // 参考：https://learnopengl-cn.readthedocs.io/zh/latest/05%20Advanced%20Lighting/04%20Normal%20Mapping/
#ifdef BUMPED
    vec3 nWorldSpace = normalize(v_tangentToWorldSpaceTransformMatrix * normalVector);
    vec3 vWorldSpace = normalize(v_cameraDirectionWorldSpace);
    vec3 reflectionWorldSpace = -normalize(reflect(vWorldSpace, nWorldSpace));
#else
    vec3 nWorldSpace = n;
    vec3 vWorldSpace = v;
    vec3 reflectionWorldSpace = -normalize(reflect(vWorldSpace, nWorldSpace));
#endif
    color += getIBLContribution(perceptualRoughness, NdotV, diffuseColor, specularColor, nWorldSpace, reflectionWorldSpace);
#else
    color += getAmbientContribution(perceptualRoughness, NdotV, diffuseColor, specularColor);
#endif

    // Apply optional PBR terms for additional (optional) shading
#ifdef HAS_AOMAP
    float ao = SRGBtoLINEAR(texture2D(u_aoMap, v_texCoord)).r;
    color = mix(color, color * ao, u_ao);
#endif

#ifdef HAS_EMISSIVEMAP
    vec3 emissive = SRGBtoLINEAR(texture2D(u_emissiveMap, v_texCoord)).rgb * u_emissiveFactor;
    color += emissive;
#endif
    
#if defined(DEBUG)
    // This section uses mix to override final color for reference app visualization
    // of various parameters in the lighting equation.
    color = mix(color, g_F, u_scaleFGDSpec.x);
    color = mix(color, vec3(g_G), u_scaleFGDSpec.y);
    color = mix(color, vec3(g_D), u_scaleFGDSpec.z);
    color = mix(color, g_specContrib, u_scaleFGDSpec.w);
    
    color = mix(color, g_diffuseContrib, u_scaleDiffBaseMR.x);
    color = mix(color, baseColor.rgb, u_scaleDiffBaseMR.y);
    color = mix(color, vec3(metallic), u_scaleDiffBaseMR.z);
    color = mix(color, vec3(perceptualRoughness), u_scaleDiffBaseMR.w);
#endif
    
    // HDR tonemapping
    color = color / (color + vec3(1.0));

    // gamma correct
    return vec4(pow(color,vec3(1.0/2.2)), baseColor.a);
}
